<?php


namespace Schaffen\ProductionOrder\ProductionOrder\Managers;


use App\RestApi\ProductionOrder\Resources\ProductionOrderResource;
use Freelabois\LaravelQuickstart\Extendables\ManipulationManager;
use Illuminate\Support\Facades\DB;
use Schaffen\ProductionOrder\ProductionOrder\Repositories\CompositeProductRepository;
use Schaffen\ProductionOrder\ProductionOrder\Repositories\ProductionOrderItemRepository;
use Schaffen\ProductionOrder\ProductionOrder\Repositories\ProductionOrderRepository;
use Schaffen\ProductionOrder\ProductionOrder\Repositories\StatusRepository;
use Schaffen\WarehouseInventory\WarehouseInventory\Services\SnapshotService;
use Schaffen\ProductionOrder\ProductionOrder\Repositories\CompositeProductHistoricRepository;
use Schaffen\Core\Product\Repositories\ProductVarietyPackRepository;


class ProductionOrderManager extends ManipulationManager
{
    protected CompositeProductRepository $composite_product_repository;
    protected ProductionOrderItemManager $production_order_item_manager;
    protected ProductionOrderItemRepository $production_order_item_repository;
    protected ProductionOrderStatusManager $production_order_status_manager;
    protected StatusRepository $status_repository;
    protected SnapshotService $snapshot_service;
    protected CompositeProductHistoricRepository $composite_product_historic_repository;

    public function __construct(ProductionOrderRepository $repository)
    {
        parent::__construct($repository);
//        dd('a');

        $this->composite_product_repository = app()->make(CompositeProductRepository::class);
        $this->production_order_item_manager = app()->make(ProductionOrderItemManager::class);
        $this->production_order_status_manager = app()->make(ProductionOrderStatusManager::class);

        $this->status_repository = app()->make(StatusRepository::class);
        $this->snapshot_service = app()->make(SnapshotService::class);


        $this->production_order_item_repository = app()->make(ProductionOrderItemRepository::class);

        $this->composite_product_historic_repository = app()->make(CompositeProductHistoricRepository::class);

        $this->product_variety_pack_repository = app()->make(ProductVarietyPackRepository::class);
    }

    public function index($filter) {
        return $this->repository->list($filter, [
            'finished_product',
            'finished_product.measurement_unit',
            'status',
        ]);
    }

    public function storeOrUpdate($values, $id = null, array $relations = [])
    {
        //gravar ordem de producao com status aguardando producao

        return DB::transaction(function () use ($values, $id, $relations) {

            $production_order = $this->repository->storeOrUpdate($values)->load(['finished_product']);

            $composition_products = $this->composite_product_repository->findMany(
                [
                    "pcp_fk_pve_codigo_acabado" => $values['op_fk_pve_final']
                ]
            );

            $composition_products = $composition_products->map(function ($product) use ($production_order) {
                $product_variety_pack = $this->product_variety_pack_repository->firstOrFail([
                    'pve_codigo' => $product->pcp_fk_pve_codigo_materia_prima
                ], ['product_variety']);

                $total_value = $product->pcp_quantidade * $product_variety_pack->product_variety->prv_valor_venda;

                return [
                    "opi_fk_pve_codigo" => $product->pcp_fk_pve_codigo_materia_prima,
                    "opi_fk_op_codigo" => $production_order->op_codigo,
                    "opi_quantidade_inicial" => $product->pcp_quantidade,
                    "opi_quantidade_final" => $product->pcp_quantidade,
                    "opi_valor_inicial" => $total_value,
                    "opi_valor_final" => $total_value
                ];
            });

            $production_order_items = [];
            foreach ($composition_products as $product) {
                $production_order_items[] = $this->production_order_item_repository->storeOrUpdate($product)->load(['feedstock']);
            }

            $this->composite_product_historic_repository->storeOP(
                $production_order,
                $production_order_items,
                CompositeProductHistoricRepository::STORE_OP
            );

            $status = $this->status_repository->firstOrFail();

            $production_order_status = [
                "ops_fk_op_codigo" => $production_order->op_codigo,
                "ops_fk_status" => $status->st_codigo
            ];

            $this->production_order_status_manager->storeOrUpdate($production_order_status);

            return new ProductionOrderResource($production_order->load('status_po', 'finished_product'));

        });


    }

    public function update($values, $id) {  
        $production_order = $this->repository->firstOrFail([
            'op_codigo' => $id
        ], [
            'status',
        ]);

        if (!$production_order->status[0]->editable()) {
            throw new \Exception("A ordem de produção não pode ser alterada no status atual");
        }

        return DB::transaction(function() use ($production_order, $values) {
            $production_order->fill($values);
            $production_order->save();

            return $production_order;
        });
    }

    public function show($id) {
        return $this->repository->firstOrFail([
            'op_codigo' => $id
        ], [
            'items'
        ]);
    }

    public function changeProductionOrderStatus(array $values, int $id)
    {

        return DB::transaction(function () use ($values, $id) {


            $production_order = $this->repository->find($id, ['status_po', 'finished_product']);
            $status = $this->status_repository->find($values['st_codigo']);

            if ($status->st_codigo == $production_order->status_po->ops_fk_status) {
                throw new \Exception("Essa ordem já se encontra no status '$status->st_descricao'");
            }

            if ($production_order->status_po->st_finalizado) {
                throw new \Exception("Essa ordem de produção já se encontra finalizada!");
            }

            if (boolval($status->st_controla_estoque)) {

                $items = $this->production_order_item_repository->findMany(["opi_fk_op_codigo" => $id]);

                $fil_codigo = $this->snapshot_service->getBranchUser()->usf_fk_fil_codigo;
                $warehouse_inventory = $this->snapshot_service->getInventoryFromBranch($fil_codigo);
                $warehouse_inventory_destiny = $this->snapshot_service->getProductionOrderWarehouse($fil_codigo);
                foreach ($items as $item) {

                    if (!boolval($status->st_finalizado)) {
                        $snapshot = $this->snapshot_service->validateProductInventory(
                            $warehouse_inventory->loe_codigo,
                            $item->opi_fk_pve_codigo,
                            $item->opi_quantidade_inicial
                        );

                        if (!is_null($snapshot)) {

                            $acc_out = $this->snapshot_service->createAccountInventory($id,
                                $item->opi_quantidade_inicial,
                                $item->opi_fk_pve_codigo,
                                $item->opi_valor_inicial,
                                $warehouse_inventory->loe_codigo,
                                0,
                                'SAIDA'
                            );

                            $acc_in = $this->snapshot_service->createAccountInventory($id,
                                $item->opi_quantidade_inicial,
                                $item->opi_fk_pve_codigo,
                                $item->opi_valor_inicial,
                                $warehouse_inventory_destiny->loe_codigo,
                                1,
                                'ENTRADA'
                            );

                        } else {
                            throw new \Exception("Produto nº {$item->opi_fk_pve_codigo} não possui estoque para a" .
                                " quantidade {$item->opi_quantidade_inicial}.");
                        }

                        $item->opi_fk_cc_codigo_origem = $acc_out->cc_codigo;
                        $item->opi_fk_cc_codigo_destino = $acc_in->cc_codigo;
                        $item->opi_fk_loe_codigo_origem = $warehouse_inventory->loe_codigo;
                        $item->opi_fk_loe_codigo_destino = $warehouse_inventory_destiny->loe_codigo;

                        $item->save();
                    } else {
                        $this->snapshot_service->createAccountInventory($id,
                            $item->opi_quantidade_final,
                            $item->opi_fk_pve_codigo,
                            $item->opi_valor_final,
                            $warehouse_inventory_destiny->loe_codigo,
                            0,
                            'SAIDA'
                        );
                    }

                }
                if (boolval($status->st_finalizado)) {
//                    ;
                    $this->snapshot_service->createAccountInventory($id, $production_order->op_quantidade,
                        $production_order->op_fk_pve_final,
                        $production_order->finished_product->product_variety->prv_custo_total,
                        $warehouse_inventory_destiny->loe_codigo,
                        1,
                        'ENTRADA'
                    );
                }
            }

            $this->production_order_status_manager->storeOrUpdate(["ops_fk_status" => $status->st_codigo],
                $production_order->status_po->ops_codigo);

            return $production_order->load('status_po');
        });
    }

}
